
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <CFNetwork/CFSocketStream.h>

#import "Server.h"
#import "Connection.h"
#import "AppConfig.h"


@interface Server ()
@property(nonatomic,assign) uint16_t port;
@property(nonatomic,strong) NSNetService* netService;

- (BOOL)createServer;
- (void)terminateServer;

- (BOOL)publishService;
- (void)unpublishService;
@end



@implementation Server

@synthesize delegate;
@synthesize port, netService;





- (BOOL)start {

  if ( ! [self createServer] ) {
    return NO;
  }
  

  if ( ! [self publishService] ) {
    [self terminateServer];
    return NO;
  }
  
  return YES;
}



- (void)stop {
  [self terminateServer];
  [self unpublishService];
}


#pragma mark Callbacks


- (void)handleNewNativeSocket:(CFSocketNativeHandle)nativeSocketHandle {
  Connection* connection = [[Connection alloc] initWithNativeSocketHandle:nativeSocketHandle];
  

  if ( connection == nil ) {
    close(nativeSocketHandle);
    return;
  }
  

  if ( ! [connection connect] ) {
    [connection close];
    return;
  }
  

  [delegate handleNewConnection:connection];
}



static void serverAcceptCallback(CFSocketRef socket, CFSocketCallBackType type, CFDataRef address, const void *data, void *info) {
  Server *server = (__bridge Server*)info;
  

  if ( type != kCFSocketAcceptCallBack ) {
    return;
  }
  

  CFSocketNativeHandle nativeSocketHandle = *(CFSocketNativeHandle*)data;

  [server handleNewNativeSocket:nativeSocketHandle];
}


#pragma mark Sockets and streams

- (BOOL)createServer {

  CFSocketContext socketCtxt = {0, (__bridge void *)(self), NULL, NULL, NULL};

  listeningSocket = CFSocketCreate(
    kCFAllocatorDefault,
    PF_INET,        
    SOCK_STREAM,    
    IPPROTO_TCP,    
    kCFSocketAcceptCallBack,  
    (CFSocketCallBack)&serverAcceptCallback,
    &socketCtxt );
   
  
  if ( listeningSocket == NULL ) {
    return NO;
  }


  int existingValue = 1;


  setsockopt( CFSocketGetNative(listeningSocket),
    SOL_SOCKET, SO_REUSEADDR, (void *)&existingValue,
    sizeof(existingValue));



  struct sockaddr_in socketAddress;
  memset(&socketAddress, 0, sizeof(socketAddress));
  socketAddress.sin_len = sizeof(socketAddress);
  socketAddress.sin_family = AF_INET;   
  socketAddress.sin_port = 0;           
  socketAddress.sin_addr.s_addr = htonl(INADDR_ANY);      

  NSData *socketAddressData =
    [NSData dataWithBytes:&socketAddress length:sizeof(socketAddress)];
  

  if ( CFSocketSetAddress(listeningSocket,
      (__bridge CFDataRef)socketAddressData) != kCFSocketSuccess ) {

    if ( listeningSocket != NULL ) {
      CFRelease(listeningSocket);
      listeningSocket = NULL;
    }
    
    return NO;
  }
  


  NSData *socketAddressActualData = 
      (NSData *)CFBridgingRelease(CFSocketCopyAddress(listeningSocket));


  struct sockaddr_in socketAddressActual;
  memcpy(&socketAddressActual, [socketAddressActualData bytes],
      [socketAddressActualData length]);
      
  self.port = ntohs(socketAddressActual.sin_port);
  

  CFRunLoopRef currentRunLoop = CFRunLoopGetCurrent();
  CFRunLoopSourceRef runLoopSource = CFSocketCreateRunLoopSource(kCFAllocatorDefault, listeningSocket, 0);
  CFRunLoopAddSource(currentRunLoop, runLoopSource, kCFRunLoopCommonModes);
  CFRelease(runLoopSource);
  
  return YES;
}


- (void) terminateServer {
  if ( listeningSocket != nil ) {
    CFSocketInvalidate(listeningSocket);
		CFRelease(listeningSocket);
		listeningSocket = nil;
  }
}


#pragma mark Bonjour

- (BOOL) publishService {

  NSString* chatRoomName = [NSString stringWithFormat:@"%@'s ChatRoom", [[AppConfig getInstance] name]];


 	self.netService = [[NSNetService alloc]
      initWithDomain:@"" type:@"_chatty._tcp."
      name:chatRoomName port:self.port];
	if (self.netService == nil)
		return NO;


	[self.netService scheduleInRunLoop:[NSRunLoop currentRunLoop]
      forMode:NSRunLoopCommonModes];


	[self.netService setDelegate:self];
  

	[self.netService publish];
  
  return YES;
}


- (void) unpublishService {
  if ( self.netService ) {
		[self.netService stop];
		[self.netService removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];
		self.netService = nil;
	}
}


#pragma mark -
#pragma mark NSNetService Delegate Method Implementations


- (void)netService:(NSNetService*)sender didNotPublish:(NSDictionary*)errorDict {
  if ( sender != self.netService ) {
    return;
  }
  

  [self terminateServer];
  

  [self unpublishService];
  

  [delegate serverFailed:self reason:@"No es pot publicar a Bonjour (nom de server incorrecte)"];
}

@end
